/*----------------------------------------------------------------------------*/
/* Copyright (c) FIRST 2008. All Rights Reserved.                             */
/* Open Source Software - may be modified and shared by FRC teams. The code   */
/* must be accompanied by the FIRST BSD license file in the root directory of */
/* the project.                                                               */
/*----------------------------------------------------------------------------*/
package edu.wpi.first.wpilibj;

import edu.wpi.first.wpilibj.communication.FRCControl;
import edu.wpi.first.wpilibj.communication.Semaphore;
import edu.wpi.first.wpilibj.communication.SemaphoreException;
import com.sun.squawk.*;

/**
 * IterativeRobot implements a specific type of Robot Program framework, extending the RobotBase class.
 *
 * The IterativeRobot class is intended to be subclassed by a user creating a robot program.
 *
 * This class is intended to implement the "old style" default code, by providing
 * the following functions which are called by the main loop, startCompetition(), at the appropriate times:
 *
 * robotInit() -- provide for initialization at robot power-on
 *
 * init() functions -- each of the following functions is called once when the
 *                     appropriate mode is entered:
 *  - DisabledInit()   -- called only when first disabled
 *  - AutonomousInit() -- called each and every time autonomous is entered from another mode
 *  - TeleopInit()     -- called each and every time teleop is entered from another mode
 *
 * Periodic() functions -- each of these functions is called iteratively at the
 *                         appropriate periodic rate (aka the "slow loop").  The default period of
 *                         the iterative robot is 10,000 microseconds, giving a periodic frequency
 *                         of 100Hz (100 times per second).
 *   - disabledPeriodic()
 *   - autonomousPeriodic()
 *   - teleopPeriodic()
 *
 * Continuous() functions -- each of these functions is called repeatedly as
 *                           fast as possible:
 *   - disabledContinuous()
 *   - autonomousContinuous()
 *   - teleopContinuous()
 *
 */
public class IterativeRobot extends RobotBase {

    private final static boolean TRACE_LOOP_ALLOCATIONS = false; // master tracing switch
    private final static boolean TRACE_LOOP_ALLOCATIONS_AFTER_INIT = true;  // trace before or after all phases initialize
    /**
     * The default period for the periodic function calls
     */
    public static final double kDefaultPeriod = .02; // default (in seconds) for periodic functions
    private boolean m_disabledInitialized;
    private boolean m_autonomousInitialized;
    private boolean m_teleopInitialized;
    private long m_disabledLoops;
    private long m_autonomousLoops;
    private long m_teleopLoops;
    private Timer m_mainLoopTimer;
    private long m_count;
    private Semaphore m_packetDataAvailableSem;

    /**
     * Constructor for RobotIterativeBase
     *
     * The constructor initializes the instance variables for the robot to indicate
     * the status of initialization for disabled, autonomous, and teleop code.
     */
    public IterativeRobot() {
        System.out.println("IterativeRobot Constructor Start");
        // set status for initialization of disabled, autonomous, and teleop code.
        m_disabledInitialized = false;
        m_autonomousInitialized = false;
        m_teleopInitialized = false;

        // keep track of the number of continuous loops performed per period
        m_disabledLoops = 0;
        m_autonomousLoops = 0;
        m_teleopLoops = 0;
        m_count = 0;

        // Start the timer for the main loop
        m_mainLoopTimer = new Timer();
        m_mainLoopTimer.start();

        Semaphore.Options options = new Semaphore.Options();
        options.setPrioritySorted(true);
        m_packetDataAvailableSem = new Semaphore(options, false);
        FRCControl.setNewDataSem(m_packetDataAvailableSem);

        System.out.println("IterativeRobot Constructor Finish");
    }

    /**
     * Provide an alternate "main loop" via startCompetition().
     *
     * This specific startCompetition() implements "main loop" behavior like that of the FRC
     * control system in 2008 and earlier, with a primary (slow) loop that is
     * called periodically, and a "fast loop" (a.k.a. "spin loop") that is
     * called as fast as possible with no delay between calls.
     */
    public void startCompetition() {
        System.out.println("IterativeRobot startCompetition() Start");
        robotInit();

        // tracing support:
        final int TRACE_LOOP_MAX = 100;
        int loopCount = TRACE_LOOP_MAX;
        Object marker = null;
        boolean didDisabledPeriodic = false;
        boolean didAutonomousPeriodic = false;
        boolean didTeleopPeriodic = false;
        if (TRACE_LOOP_ALLOCATIONS) {
            GC.initHeapStats();
            if (!TRACE_LOOP_ALLOCATIONS_AFTER_INIT) {
                System.out.println("=== Trace allocation in competition loop! ====");
                marker = new Object(); // start counting objects before any loop starts - includes class initialization
            }
        }

        // loop forever, calling the appropriate mode-dependent function
        while (true) {
            if (TRACE_LOOP_ALLOCATIONS && didDisabledPeriodic && didAutonomousPeriodic && didTeleopPeriodic && --loopCount <= 0) {
                System.out.println("!!!!! Stop loop!");
                break;
            }
            // Call the appropriate function depending upon the current robot mode
            if (isDisabled()) {
                //System.out.println("IsDisabled");
                // call DisabledInit() if we are now just entering disabled mode from
                // either a different mode or from power-on
                if (!m_disabledInitialized) {
                    disabledInit();
                    m_disabledInitialized = true;
                    // reset the initialization flags for the other modes
                    m_autonomousInitialized = false;
                    m_teleopInitialized = false;
                    System.out.println("Disabled_Init() completed");
                }
                if (nextPeriodReady()) {
                    disabledPeriodic();
                    didDisabledPeriodic = true;
                }
                disabledContinuous();
            } else if (isAutonomous()) {
                //System.out.println("IsAutonomous");

                // call Autonomous_Init() if this is the first time
                // we've entered autonomous_mode
                if (!m_autonomousInitialized) {
                    // KBS NOTE: old code reset all PWMs and relays to "safe values"
                    // whenever entering autonomous mode, before calling
                    // "Autonomous_Init()"
                    autonomousInit();
                    m_autonomousInitialized = true;
                    m_teleopInitialized = false;
                    m_disabledInitialized = false;
                    System.out.println("Autonomous_Init() completed");
                }
                if (nextPeriodReady()) {
                    getWatchdog().feed();
                    autonomousPeriodic();
                    didAutonomousPeriodic = true;
                }
                autonomousContinuous();
            } else {
                //System.out.println("Teleop");
                // call Teleop_Init() if this is the first time
                // we've entered teleop_mode
                if (!m_teleopInitialized) {
                    teleopInit();
                    m_teleopInitialized = true;
                    m_autonomousInitialized = false;
                    m_disabledInitialized = false;
                    System.out.println("Teleop_Init() completed");
                }
                if (nextPeriodReady()) {
                    getWatchdog().feed();
                    teleopPeriodic();
                    didTeleopPeriodic = true;
                }
                teleopContinuous();
            }
            if (TRACE_LOOP_ALLOCATIONS && TRACE_LOOP_ALLOCATIONS_AFTER_INIT &&
                    didDisabledPeriodic && didAutonomousPeriodic && didTeleopPeriodic && loopCount == TRACE_LOOP_MAX) {
                System.out.println("=== Trace allocation in competition loop! ====");
                marker = new Object(); // start counting objects after 1st loop completes - ignore class initialization
            }
        }
        if (TRACE_LOOP_ALLOCATIONS) {
            GC.printHeapStats(marker, false);
        }
    }

    /**
     * Determine if the appropriate next periodic function should be called.
     *
     * This function makes adjust for lateness in the cycle by keeping track of how late
     * it was and crediting the next period with that amount.
     *
     * @todo Decide what this should do if it slips more than one cycle.
     */
    private boolean nextPeriodReady() {
        m_count++;
        try {
            if (m_packetDataAvailableSem.tryTake()) {
                double curTime = m_mainLoopTimer.get();
                m_mainLoopTimer.reset();
                m_count = 0;
                return true;
            }
            return false;
        } catch (SemaphoreException e) {
            return false;
        }
    }

//    private boolean nextPeriodReady() {
//        m_count++;
//        try {
//            m_packetDataAvailableSem.take(0);
//            if (m_mainLoopTimer.get() > m_period * 2.0) {
//                System.out.println("P " + (m_mainLoopTimer.get() - m_period) + " " + m_count + "\n");
//            }
//            m_mainLoopTimer.reset();
//            m_count = 0;
//            return true;
//        } catch (SemaphoreException e) {
//        }
//        return false;
//    }
    /**
     * Robot-wide initialization code should go here.
     *
     * Users should override this method for default Robot-wide initialization which will
     * be called when the robot is first powered on.  It will be called exactly 1 time.
     */
    public void robotInit() {
        System.out.println("Default RobotIterativeBase::RobotInit() method running");
    }

    /**
     * Initialization code for disabled mode should go here.
     *
     * Users should override this method for initialization code which will be called each time
     * the robot enters disabled mode.
     */
    public void disabledInit() {
        System.out.println("Default RobotIterativeBase::DisabledInit() method running");
    }

    /**
     * Initialization code for autonomous mode should go here.
     *
     * Users should override this method for initialization code which will be called each time
     * the robot enters autonomous mode.
     */
    public void autonomousInit() {
        System.out.println("Default RobotIterativeBase::AutonomousInit() method running");
    }

    /**
     * Initialization code for teleop mode should go here.
     *
     * Users should override this method for initialization code which will be called each time
     * the robot enters teleop mode.
     */
    public void teleopInit() {
        System.out.println("Default RobotIterativeBase::TeleopInit() method running");
    }

    /**
     * Periodic code for disabled mode should go here.
     *
     * Users should override this method for code which will be called periodically at a regular
     * rate while the robot is in disabled mode.
     */
    public void disabledPeriodic() {
        System.out.print("D - ");
        System.out.println(m_disabledLoops);
        m_disabledLoops = 0;
    }

    /**
     * Periodic code for autonomous mode should go here.
     *
     * Users should override this method for code which will be called periodically at a regular
     * rate while the robot is in autonomous mode.
     */
    public void autonomousPeriodic() {
        System.out.print("A - ");
        System.out.println(m_autonomousLoops);
        m_autonomousLoops = 0;
    }

    /**
     * Periodic code for teleop mode should go here.
     *
     * Users should override this method for code which will be called periodically at a regular
     * rate while the robot is in teleop mode.
     */
    public void teleopPeriodic() {
        System.out.print("T - ");
        System.out.println(m_teleopLoops);
        m_teleopLoops = 0;
    }

    /**
     * Continuous code for disabled mode should go here.
     *
     * Users should override this method for code which will be called repeatedly as frequently
     * as possible while the robot is in disabled mode.
     */
    public void disabledContinuous() {
        m_disabledLoops++;
    }

    /**
     * Continuous code for autonomous mode should go here.
     *
     * Users should override this method for code which will be called repeatedly as frequently
     * as possible while the robot is in autonomous mode.
     */
    public void autonomousContinuous() {
        m_autonomousLoops++;
    }

    /**
     * Continuous code for teleop mode should go here.
     *
     * Users should override this method for code which will be called repeatedly as frequently
     * as possible while the robot is in teleop mode.
     */
    public void teleopContinuous() {
        m_teleopLoops++;
    }
}
