/*----------------------------------------------------------------------------*/
/* Copyright (c) FIRST 2008. All Rights Reserved.                             */
/* Open Source Software - may be modified and shared by FRC teams. The code   */
/* must be accompanied by the FIRST BSD license file in the root directory of */
/* the project.                                                               */
/*----------------------------------------------------------------------------*/

package edu.wpi.first.wpilibj;

import javax.microedition.midlet.MIDlet;
import javax.microedition.midlet.MIDletStateChangeException;

/**
 * Implement a Robot Program framework.
 * The RobotBase class is intended to be subclassed by a user creating a robot program.
 * Overridden autonomous() and operatorControl() methods are called at the appropriate time
 * as the match proceeds. In the current implementation, the Autonomous code will run to
 * completion before the OperatorControl code could start. In the future the Autonomous code
 * might be spawned as a task, then killed at the end of the Autonomous period.
 */
public abstract class RobotBase extends MIDlet {

    private final DriverStation m_ds;
    private final Watchdog m_watchdog = Watchdog.getInstance();
    private int previousPacketNumber;

    /**
     * Constructor for a generic robot program.
     * User code should be placed in the constuctor that runs before the Autonomous or Operator
     * Control period starts. The constructor will run to completion before Autonomous is entered.
     *
     * This must be used to ensure that the communications code starts. In the future it would be
     * nice to put this code into it's own task that loads on boot so ensure that it runs.
     */
    protected RobotBase() {
        // TODO: StartCAPI();
        // TODO: See if the next line is necessary
        // Resource.RestartProgram();
        m_ds = DriverStation.getInstance();
    }

    /**
     * Free the resources for a RobotBase class.
     */
    protected void free() {
    }

    /**
     * Check on the overall status of the system.
     *
     * @return Is the system active (i.e. PWM motor outputs, etc. enabled)?
     */
    public boolean isSystemActive() {
        return m_watchdog.isSystemActive();
    }

    /**
     * Return the instance of the Watchdog timer.
     * Get the watchdog timer so the user program can either disable it or feed it when
     * necessary.
     *
     * @return The Watchdog timer.
     */
    public Watchdog getWatchdog() {
        return m_watchdog;
    }

    /**
     * Determine if the Robot is currently disabled.
     * @return True if the Robot is currently disabled by the field controls.
     */
    public boolean isDisabled() {
        return m_ds.isDisabled();
    }

    /**
     * Determine if the Robot is currently enabled.
     * @return True if the Robot is currently enabled by the field controls.
     */
    public boolean isEnabled() {
        return m_ds.isEnabled();
    }

    /**
     * Determine if the robot is currently in Autnomous mode.
     * @return True if the robot is currently operating Autonomously as determined by the field controls.
     */
    public boolean isAutonomous() {
        return m_ds.isAutonomous();
    }

    /**
     * Determine if the robot is currently in Operator Control mode.
     * @return True if the robot is currently operating in Tele-Op mode as determined by the field controls.
     */
    public boolean isOperatorControl() {
        return m_ds.isOperatorControl();
    }

    /**
     * Indicates if new data is available from the driver station.
     * @todo The current implementation is silly.  We already know this explicitly without trying to figure it out.
     * @return Has new data arrived over the network since the last time this function was called?
     */
    public boolean isNewDataAvailable() {
        if (m_ds.getPacketNumber() == previousPacketNumber) {
            return false;
        }

        previousPacketNumber = m_ds.getPacketNumber();
        return true;
    }

    /**
     * Provide an alternate "main loop" via startCompetition().
     */
    public abstract void startCompetition();

    /**
     * Starting point for the applications. Starts the OtaServer and then runs
     * the robot.
     * @throws javax.microedition.midlet.MIDletStateChangeException
     */
    protected final void startApp() throws MIDletStateChangeException {
//        OtaCommandServerSocketListener.startListener();
//        PipeServerSocketListener.startPipingConsoleIfNecessary();

        Watchdog.getInstance().setExpiration(0.1);
        Watchdog.getInstance().setEnabled(true);

        try {
            this.startCompetition();
        } catch (Throwable t) {
            t.printStackTrace();
        }

        // startCompetition never returns unless exception occurs....
        // TODO: work out what to do here - should not need to do anything
        // once we have two Squawk processes?
        // For now, don't quit, so the OTA server stays running

        for (;;) {
            try {
                Thread.sleep(10000);
            } catch (InterruptedException ex) {
                ex.printStackTrace();
            }
        }
    }

    /**
     * Pauses the application
     */
    protected final void pauseApp() {
        // This is not currently called by the Squawk VM
    }

    /**
     * Called if the MIDlet is terminated by the system.
     * I.e. if startApp throws any exception other than MIDletStateChangeException,
     * if the isolate running the MIDlet is killed with Isolate.exit(), or
     * if VM.stopVM() is called.
     *
     * It is not called if MIDlet.notifyDestroyed() was called.
     *
     * @param unconditional If true when this method is called, the MIDlet must
     *    cleanup and release all resources. If false the MIDlet may throw
     *    MIDletStateChangeException  to indicate it does not want to be destroyed
     *    at this time.
     * @throws MIDletStateChangeException if there is an exception in terminating the midlet
     */
    protected final void destroyApp(boolean unconditional) throws MIDletStateChangeException {
    }
}
